{{
        Karplus-Strong on the propeller
        (sounds like a plucked string)

        Jonathan "lonesock" Dummer

        NOTES:

        * use any buffer size you want
          - no power-of-2-restriction
          - buffer uses words

        * fraction_bits must be [1..8]
          - more means more accurate delay-sample interpolation
          - more means a bit more computation (12-clocks per bit)

        * the period of the plucked string is specified in fixed-point
          - if 'fraction_bits' = 4, then you should use period * 2^4,
            and the code will use the period to the nearest 1/16th of
            a sample.
          - the exact period isn't precisely determined yet..I still
            have to do some math regarding the simple filtering I used.

        * you can change the sustain of the string with 'sustain'
          - implemented as scaling the feedback by (2^N - 1) / (2^N)
          - the higher this number...the longer it will ring
          - set once in start...no changing without restarting

        * pluck_note uses the MIDI note convention (69 => A440)

        TODO:

        * allow changing the decay dynamically

        * simulate multiple strings in a single cog
}}
CON
  buf_size      = 2345          ' any size you want...just need enough RAM
  fraction_bits = 8             ' must be in the range [1..8]

  ' used in my control array
  ctrl_VOLUME   = 0             ' set this to [1..31] to "pluck" a string
  ctrl_DELAY    = 1             ' set this to the period * 2^fraction_bits to set the frequency
  ctrl_SUSTAIN  = 2             ' set this to [1..31] to control how long the string rings

DAT
{{
  Matched to the MIDI spec, freq = 440 * 2 ^ ( (N - 69) / 12 )
  MIDI note ratios for 0..11 (every 12 notes goes up an octave...just divide the period by 2)
  fixed point period = sample_rate << fraction_bits / frequency
  period = sample_rate << fraction_bits * 2^32 / 2^32 / frequency
  using the ** operator (x ** y) means (x * y / 2^32)
  so if ratio = 2^32 / frequency
  period = sample_rate << fraction_bits ** ratio
}}
ratio0  long $1F4FDA4C  ' 2^32 / 8.175799 Hz
ratio1  long $1D8DF51D  ' 2^32 / 8.661957 Hz
ratio2  long $1BE5501B  ' 2^32 / 9.177024 Hz
ratio3  long $1A548077  ' 2^32 / 9.722718 Hz
ratio4  long $18DA2FC0  ' 2^32 / 10.300861 Hz
ratio5  long $17751ABC  ' 2^32 / 10.913382 Hz
ratio6  long $16241056  ' 2^32 / 11.562326 Hz
ratio7  long $14E5F099  ' 2^32 / 12.249857 Hz
ratio8  long $13B9ABB9  ' 2^32 / 12.978272 Hz
ratio9  long $129E412A  ' 2^32 / 13.750000 Hz
ratio10 long $1192BEC8  ' 2^32 / 14.567618 Hz
ratio11 long $10964004  ' 2^32 / 15.433853 Hz

VAR
  long  control[3]              ' here's where I pass commands to the PASM cog (PAR)
  long  sample_rate_fixed       ' sample-rate << fraction_bits, used in pluck_note
  word  buf[buf_size]           ' using words to minimize the memory cost
  byte  cog                     ' which cog (if any) am I using?

{ simple test code...remove the leading ' to comment this block out
CON
  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000
  sr            = 44_100        ' let's run at the CD sampling rate
  sustain_value = 8
  demo_pinL     = 10
  demo_pinR     = 11
OBJ
  term          : "FullDuplexSerial"
PUB test_KS | rn, p

  ' set up my serial port, and wait for the BST terminal to go live
  term.start( 31, 30, 0, 115_200 )

  ' set up using the demo board pins for the headphones
  start( demo_pinL, demo_pinR, sr )

  waitcnt( clkfreq >> 3 + cnt )

  ' play the scale
  rn := 0
  set_sustain( 10 )
  repeat
    ' wait a bit
    waitcnt( clkfreq / 4 + cnt )
    ' pluck a note (MIDI spec) and returns the period
    set_note( rn )
    if( rn & %11 == 0 )
      pluck( 31 )
    ' debug
    term.dec( rn )
    term.tx( "," )
    term.dec( p )
    term.tx( 13 )
    ' next note?
    ++rn
    rn //= 128
    { do a quick loop around the doubler transition point
    if rn > 50
      rn := 30
      waitcnt( clkfreq * 3 / 4 + cnt )
    '}
'}

{{
  The minimum period is < 500 clocks, so keep the sample rate < clock / 500
  So, for the prop running at 80MHz, the sample rate should be less that 160 kHz [8^)
}}
PUB start( pinL, pinR, sample_rate )
  stop
  ' set up the counter and pins
  if (pinL < 0) AND (pinR < 0)
    return 0
  if (pinL > -1) AND (pinR > -1)
    ctra_val := constant(%00111 << 26) | (pinL << 9) | pinR
    control := |<pinL | |<pinR
  else
    if pinL > -1
      ctra_val := constant(%00110 << 26) | pinL
      control := |<pinL
    else
      ctra_val := constant(%00110 << 26) | pinR
      control := |<pinR
  ' set up the pseudo-constants used in the PASM code
  sample_rate_fixed := sample_rate << fraction_bits
  period := (clkfreq + (sample_rate >> 1)) / sample_rate
  buf_head := @buf
  buf_tail := buf_head + constant(buf_size * 2)
  max_delay := constant( (buf_size - 2) << fraction_bits )
  ' start it up
  control[ctrl_DELAY] := constant( 1 << fraction_bits )
  control[ctrl_SUSTAIN] := 8
  cog := cognew( @ks_PASM, @control ) + 1
  return cog

PUB set_sustain( sustain )
  control[ctrl_SUSTAIN] := sustain #> 0 <# 31

PUB set_frequency( Hz )
  control[ctrl_DELAY] := sample_rate_fixed / Hz

PUB set_note( MIDI_note ) : p
  control[ctrl_DELAY] := sample_rate_fixed ** ratio0[MIDI_note//12] >> (MIDI_note / 12)

PUB pluck( volume )
  control := volume #> 0 <# 31

PUB stop
  if cog
    cogstop( cog~ - 1 )

DAT
ORG 0

ks_PASM       ' setup
              mov       frqa, middle_val
              mov       ctra, ctra_val
              rdlong    dira, par
              mov       timestamp, period
              add       timestamp, cnt
              wrlong    phsb, par
              mov       ptr_write, buf_head
              mov       ptr_read, buf_head

loop          ' main loop: wait for the proper time, then after a constant delay set the output value (no rd*/wr* yet!)
              waitcnt   timestamp, period
              add       sample, middle_val
              mov       frqa, sample

              ' OK, what are my operational parameters?
              mov       tmp, par
              add       tmp, #8
              rdlong    sustain_sar, tmp
              sub       tmp, #4
              rdlong    int_delay, tmp

              ' update those settings, whether they changed or not
              call      #adjust_delay

              ' does the user want to pluck the string?
              rdlong    tmp, par wz
        if_nz mov       pluck_sar, #%11111
        if_nz xor       pluck_sar, tmp
        if_nz wrlong    phsb, par
        if_nz mov       pluck_left, int_delay

              ' generate a pseudo-random number as a white-noise input.
              add       rnd_seed, prng_add      ' $80808080
              ror       rnd_seed, #17
              mov       sample, rnd_seed
              sar       sample, pluck_sar

              ' perform the delay / filter / feedback
              mov       last_val, this_val
              add       ptr_read, #2
              cmpsub    ptr_read, buf_tail wc
        if_c  add       ptr_read, buf_head
              ' read this value and scale
              rdword    this_val, ptr_read
              shl       this_val, #16

{{
              This section does a linear interpolation.
              BUT
              We need filtered samples to sound OK, so we first do a simple
              average-of-2-sample lowpass, then we interpolate between 2 of
              those filtered samples!
}}
              ' shuffle the most recent filtered value into it's correct place
              mov       last_filter, this_filter
              ' and calculate the newest filtered value
              mov       this_filter, this_val
              sar       this_filter, #1
              mov       tmp, last_val
              sar       tmp, #1
              add       this_filter, tmp

              '{ interpolation of the filtered values
              mov       tmp, this_filter        ' starting point
              mov       tmpt, last_filter
              sar       tmp, #1
              sar       tmpt, #1
              sub       tmpt, tmp               ' the delta
              shr       tmpt, #fraction_bits    ' place the fractional bits
              shl       tmpt, #fraction_bits    ' (already in reverse order)
              or        tmpt, interp_frac       ' into the low 4 bits of tempt
              mov       tmpc, #fraction_bits                ' 4 fractional bits
              ' do the interpolation
:frac_loop    sar       tmpt, #1 wc
        if_c  add       tmp, tmpt
              djnz      tmpc, #:frac_loop

              add       tmp, tmp                ' get the scale correct again
              '}
              { alternatively: just do simple averaging: no sub-sample delay resolution
              mov       tmp, last_val
              mov       tmpt, this_val
              sar       tmp, #1
              sar       tmpt, #1
              add       tmp, tmpt
              '}

              ' Are we plucking?  Just leave it random
              cmpsub    pluck_left, #1 wc
              ' otherwise use the value for sampled from the filtered delay line
        if_nc mov       sample, tmp
              ' but now decay it a bit
        if_nc sar       tmp, sustain_sar
        if_nc sub       sample, tmp

              ' write this sample, and advance
              mov       tmp, sample
              sar       tmp, #16
              wrword    tmp, ptr_write
              add       ptr_write, #2
              cmpsub    ptr_write, buf_tail wc
        if_c  add       ptr_write, buf_head

              ' loop
              jmp       #loop

adjust_delay
              ' adjust the input period to handle the filter offset
              add       int_delay, #1 << fraction_bits >> 1
              ' and make sure we don't exceed our limit
              max       int_delay, max_delay
              ' calculate the interpolation fraction (next "fraction_bits" bits)
              mov       interp_frac, int_delay
              rev       interp_frac, #32 - fraction_bits
              shr       int_delay, #fraction_bits
              ' and finally recover the delay (address in words...2 bytes per sample)
              ' (leave ptr_write alone, just change ptr_read)
              mov       ptr_read, ptr_write
              add       ptr_read, buf_tail
              sub       ptr_read, buf_head
              sub       ptr_read, int_delay
              sub       ptr_read, int_delay
              ' wrap it as necessary
:wrap         cmpsub    ptr_read, buf_tail wc
        if_c  add       ptr_read, buf_head
        if_c  jmp       #:wrap
              ' done!
adjust_delay_ret
              ret


{=== PASM faux-constants, or initialized variables ===}
prng_add      long      $80808080
middle_val    long      |<31
sample        long      0
pluck_left    long      0
pluck_sar     long      31
interp_frac   long      0
last_filter   long      0
this_filter   long      0


{=== PASM parameters ===}
ctra_val      long      0
period        long      0
buf_head      long      0
buf_tail      long      0
max_delay     long      0
this_val      long      0
last_val      long      0

{=== PASM scratch variables ===}
sustain_sar   res       1       ' means scale the feedback by (2^N - 1) / (2^N)
rnd_seed      res       1
tmp           res       1
tmpl          res       1
tmpt          res       1
tmpc          res       1
timestamp     res       1
int_delay     res       1
ptr_write     res       1
ptr_read      res       1
